/*
 * mutex_ex.cpp
 *
 *  Created on: Mar 24, 2011
 *      Author: lipari
 */

#include <boost/thread.hpp>
#include <boost/ref.hpp>
#include <iostream>
#include <string>
#include "example_map.hpp"

#include "mutex_ex.hpp"

namespace example_mutex_cond
{
	oosd::register_in_map __reg("mutex_example", run);

	using namespace std;

	class CircularArray {
		boost::mutex m;
		boost::condition_variable empty;
		boost::condition_variable full;
		int head;
		int tail;
		int num;
		int max;
		int *array;
	public:
		CircularArray(int n) : head(0), tail(0), num(0), max(n) {
			array = new int[n];
		}
		~CircularArray() {
			delete [] array;
		}

		bool isFull() { return num==max; }
		bool isEmpty() { return num==0; }

		void insert(int elem) {
			boost::unique_lock<boost::mutex> guard(m);

			while (isFull()) full.wait(guard);
			array[head] = elem;
			head = (head +1) % max;
			num++;
			empty.notify_one();
		}

		int extract() {
			boost::unique_lock<boost::mutex> guard(m);
			while(isEmpty()) empty.wait(guard);
			int elem = array[tail];
			tail = (tail+1) % max;
			num--;
			full.notify_one();
			return elem;
		}
	};

	void producer_body(CircularArray &ca)
	{
		for (int i=0; i<10; i++) {
			cout << "Inserting : " << i << endl;
			ca.insert(i);
			cout << "Inserted  : " << i << endl;
		}
	}

	void consumer_body(CircularArray &ca)
	{
		for (int i=0; i<10; i++) {
			cout << "Extracting" << endl;
			int k = ca.extract();
			cout << "Extracted : " << k << endl;
		}
	}

	int run() {
		CircularArray ca(4);
		boost::thread prod(producer_body, boost::ref(ca));
		boost::thread cons(consumer_body, boost::ref(ca));

		prod.join();
		cons.join();
		cout << "end" << endl;
		return 0;
	}
}
